//
//  AlliedCamera.cpp
//  emptyExample
//
//  Created by Theodore Watson on 9/30/14.
//
//

#include "AlliedCamera.h"
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include <stdio.h>


#define MAX_CAMERA_LIST 50
static int initCount = 0;
static int maxNumberCameras = 6;
static int bHasLinkListeners = false;

AlliedPvLinkListener listener;
ofEvent< AlliedLinkEvent > AlliedPvLinkListener::LinkEvent;

#pragma mark - Public API

//------------------------------------------------------------------------
AlliedCamera::AlliedCamera(){
    reset();
}

//------------------------------------------------------------------------
AlliedCamera::~AlliedCamera(){
    cout << "~AlliedCamera()" << endl;
    closeCamera();
}

//------------------------------------------------------------------------
bool AlliedCamera::openCamera(unsigned int deviceID, bool waitForFirstFrame){
    closeCamera();
    
    vector <CameraInfo> cameras = AlliedCamera::ListDevices();
    if( deviceID <= cameras.size() ){
        info = cameras[deviceID];
    }else{
        ofLogError("AlliedCamera::openCamera") << " device " << deviceID << " to large. Only " << cameras.size() << " devices found " << endl;
        return false;
    }
    
    if( info.available == false || info.ipAddress == "" ){
        ofLogError("AlliedCamera::openCamera") << " device " << deviceID << " " << info.deviceName << " does not have an IP address or is not available " << endl;
        reset();
        return false;
    }
    
    ofLogNotice("AlliedCamera::openCamera") << " opening camera" << endl; 
    startThreadedCapture(waitForFirstFrame);

    return true;
}

//------------------------------------------------------------------------
bool AlliedCamera::openCamera(string ipAddress, bool waitForFirstFrame){
    closeCamera();
    
    vector <CameraInfo> cameras = AlliedCamera::ListDevices();
    bool bFound = false;
    
    for(int i = 0; i < cameras.size(); i++){
        if(cameras[i].ipAddress == ipAddress){
            info = cameras[i];
            bFound = true;
            break;
        }
    }
    
    if( bFound ){
        if( info.available == false || info.ipAddress == "" ){
            ofLogError("AlliedCamera::openCamera") << " device " << info.deviceNo << " " << info.deviceName << " does not have an IP address or is not available " << endl;
            reset();
            return false;
        }
    }else{
        ofLogError("AlliedCamera::openCamera") << " device with IP address: " << ipAddress << " not found " << endl;
        return false;
    }
    
    startThreadedCapture(waitForFirstFrame);
    
    return true;
}

//--------------------------------------------------------------
vector <AlliedCamera::CameraParam> AlliedCamera::getCameraParameters(){
    if( isThreadRunning() ){
        buildAttributeList();
    }

    return camParams;
}

//--------------------------------------------------------------
void AlliedCamera::closeCamera(){
    if( isThreadRunning() ){
        waitForThread(true);
        releaseApi();
    }
    reset();
}

//--------------------------------------------------------------
bool AlliedCamera::isCameraRunning(){
    return isThreadRunning();
}

//--------------------------------------------------------------
void AlliedCamera::update(){
    if( bFrameNewInteral && isThreadRunning() ){
        if( lock() ){
            pixels = pixelsInternal;
            bFrameNewInteral = false;
            unlock();
        }
        bFrameNew = true;
    }else{
        bFrameNew = false;
    }
}

//--------------------------------------------------------------
bool AlliedCamera::isFrameNew(){
    return bFrameNew;
}

//--------------------------------------------------------------
const ofPixels & AlliedCamera::getPixelsRef(){
    return pixels; 
}

//--------------------------------------------------------------
bool AlliedCamera::setValueF(string paramName, float value){
    if( isThreadRunning() == false || Handle == NULL ) return false;
    
    string valString = ofToString(value);
    return String2Value(Handle, paramName, ePvDatatypeFloat32, valString);
}

//--------------------------------------------------------------
bool AlliedCamera::setValueUL(string paramName, unsigned long value){
    if( isThreadRunning() == false || Handle == NULL ) return false;
    
    string valString = ofToString(value);
    return String2Value(Handle, paramName, ePvDatatypeUint32, valString);
}

//--------------------------------------------------------------
bool AlliedCamera::setValueB(string paramName, bool value){
    if( isThreadRunning() == false || Handle == NULL ) return false;
    
    string valString = "false";
    if( value ){
        valString = "true";
    }
    return String2Value(Handle, paramName, ePvDatatypeBoolean, valString);
}

//--------------------------------------------------------------
bool AlliedCamera::setValueS(string paramName, string value){
    if( isThreadRunning() == false || Handle == NULL ) return false;
    
    return String2Value(Handle, paramName, ePvDatatypeString, value);
}

//--------------------------------------------------------------
bool AlliedCamera::setValueEnum(string paramName, string value){
    if( isThreadRunning() == false || Handle == NULL ) return false;
    
    return String2Value(Handle, paramName, ePvDatatypeEnum, value);
}

//--------------------------------------------------------------
bool AlliedCamera::getValueF(string paramName, float & value){
    if( isThreadRunning() == false || Handle == NULL ) return false;

    string valString;
    if( Value2String(Handle, paramName, ePvDatatypeFloat32, valString) ){
        value = ofToFloat(valString);
        return true;
    }
    return false;
}

//--------------------------------------------------------------
bool AlliedCamera::getValueUL(string paramName, unsigned long & value){
    if( isThreadRunning() == false || Handle == NULL ) return false;

    string valString;
    if( Value2String(Handle, paramName, ePvDatatypeUint32, valString) ){
        istringstream cur(valString);
        cur >> value;
        return true;
    }
    return false;
}

//--------------------------------------------------------------
bool AlliedCamera::getValueB(string paramName, bool & value){
    if( isThreadRunning() == false || Handle == NULL ) return false;

    string valString;
    if( Value2String(Handle, paramName, ePvDatatypeBoolean, valString) ){
        value = ( valString == "true" );
        return true;
    }
    return false;
}

//--------------------------------------------------------------
bool AlliedCamera::getValueS(string paramName, string & value){
    if( isThreadRunning() == false || Handle == NULL ) return false;

    return Value2String(Handle, paramName, ePvDatatypeString, value);
}

//--------------------------------------------------------------
bool AlliedCamera::getValueEnum(string paramName, string & value){
    if( isThreadRunning() == false || Handle == NULL ) return false;

    return Value2String(Handle, paramName, ePvDatatypeEnum, value);
}

#pragma mark - Protected API

//--------------------------------------------------------------
void AlliedCamera::startThreadedCapture(bool bWaitForFirstFrame){

    initApi();
    startThread(false);
    
    if(bWaitForFirstFrame){
        float startTime = ofGetElapsedTimef();
        float timeout = 10.0;
        while( isFrameNew() == false ){
            update();
            ofSleepMillis(10);
            if( ofGetElapsedTimef() - startTime > timeout ){
                ofLogWarning() << " AlliedCamera::openCamera - waiting for camera to open timed out";
                break;
            }
        }
    }
}

//--------------------------------------------------------------
void AlliedCamera::reset(){
    Discarded = 0;
    bFrameNewInteral = false;
    bFrameNew = false;
    info = AlliedCamera::CameraInfo();
    
    for(int i=0;i<Frames.size();i++){
        if( Frames[i]->ImageBuffer != NULL ){
            delete [] (char*)Frames[i]->ImageBuffer;
        }
        delete Frames[i];
    }
    Frames.clear();
    
    Handle = NULL;
    pixels.clear();
    pixelsInternal.clear();
    params.clear();
    for(int i = 0; i < storedParams.size(); i++){
        delete storedParams[i];
    }
    storedParams.clear();
    camParams.clear(); 
}


bool AlliedCamera::buildAttributeList(){
    if( Handle ){

        tPvUint32 lCount;
        tPvAttrListPtr lAttrs;
        
//        for(int i = 0; i < storedParams.size(); i++){
//            delete storedParams[i];
//        }
//        storedParams.clear();
//        params.clear();
//        params.setName(info.deviceName + " " + info.serialNo );
        
        camParams.clear();
        
        //Get all attributes
        if(PvAttrList(Handle,&lAttrs,&lCount) == ePvErrSuccess)
        {
            //Get info and display each attribute
            for(tPvUint32 i=0;i<lCount;i++){
                QueryAttribute(lAttrs[i]);
            }
            
            ofLogNotice() << " built param list of size " <<  camParams.size();
            
            return true;
        }
    }
    return false;
}

// display info on a given attribute of the camera
void AlliedCamera::QueryAttribute(const char* aLabel){

    tPvAttributeInfo lInfo;

    if(PvAttrInfo(Handle,aLabel,&lInfo) == ePvErrSuccess)
    {
        char lFlags[5];

        memset(lFlags,' ',sizeof(char) * 4);

        if(lInfo.Flags & ePvFlagRead)
            lFlags[0] = 'r';
        if(lInfo.Flags & ePvFlagWrite)
            lFlags[1] = 'w';
        if(lInfo.Flags & ePvFlagVolatile)
            lFlags[2] = 'v';
        if(lInfo.Flags & ePvFlagConst)
            lFlags[3] = 'c';
        lFlags[4] = '\0';
        
        CameraParam tmpParam;
        tmpParam.name = aLabel;
        tmpParam.writable = ( lFlags[1] == 'w' );
        tmpParam.type = lInfo.Datatype;
        
        if( lFlags[1] == 'w' ){
            if(lInfo.Datatype == ePvDatatypeString){
                char lValue[512];
                if(PvAttrStringGet(Handle, aLabel, lValue, 512, NULL) == ePvErrSuccess){
//                    ofParameter <string> * param = new ofParameter<string>();
//                    param->set(aLabel, lValue);
//                    params.add(*param);
//                    storedParams.push_back(param);


                       tmpParam.valueS = lValue;
                       camParams.push_back(tmpParam);
                }
            }
            else if(lInfo.Datatype == ePvDatatypeEnum){
                char lValue[512];
                if(PvAttrEnumGet(Handle, aLabel, lValue, 512, NULL) == ePvErrSuccess){
//                    ofParameter <string> * param = new ofParameter<string>();
//                    param->set(aLabel, lValue);
//                    params.add(*param);
//                    storedParams.push_back(param);

                      char enumString[2048];
                      memset(&enumString[0], 0, 2048);
                      unsigned long pSize = 0;
                      PvAttrRangeEnum(Handle, aLabel, enumString, 2048, &pSize);
                    
                      tmpParam.valueS = lValue;
            
                      vector <string> enums = ofSplitString(string(enumString), ",", true, true);
                      if( enums.size() ){
                        for(int i = 0; i < enums.size(); i++){
                            if( enums[i] == tmpParam.valueS ){
                                tmpParam.enumSelected = i;
                            }
                        }
                      }
                    
                       tmpParam.enumVals = enums;
                       camParams.push_back(tmpParam);
                }
            }
            else
            if(lInfo.Datatype == ePvDatatypeUint32){
                unsigned long lValue = 0;
                if(PvAttrUint32Get(Handle, aLabel, &lValue) == ePvErrSuccess){
//                    ofParameter <int> * param = new ofParameter<int>();
//                    param->set(aLabel, lValue);
                    
                    //try and get the range;
                    unsigned long lMin = 0;
                    unsigned long lMax = numeric_limits<int>::max();
                    PvAttrRangeUint32(Handle, aLabel, &lMin, &lMax);
                    
//                    param->setMin(lMin);
//                    param->setMax(lMax);
                    
                    tmpParam.minL = lMin;
                    tmpParam.maxL = lMax;
                    tmpParam.valueL = lValue;
                    camParams.push_back(tmpParam);

//                    params.add(*param);
//                    storedParams.push_back(param);
                }
            }
            else if(lInfo.Datatype == ePvDatatypeFloat32){
                float fValue = 0.0;
                if(PvAttrFloat32Get(Handle, aLabel, &fValue) == ePvErrSuccess){
//                    ofParameter <float> * param = new ofParameter<float>();
//                    param->set(aLabel, lValue);
                    
                    //try and get the range;
                    float fMin = 0.0;
                    float fMax = numeric_limits<float>::max();
                    PvAttrRangeFloat32(Handle, aLabel, &fMin, &fMax);
                    
//                    param->setMin(lMin);
//                    param->setMax(lMax);
                    
                    tmpParam.minF = fMin;
                    tmpParam.maxF = fMax;
                    tmpParam.valueF = fValue;
                    camParams.push_back(tmpParam);

//                    params.add(*param);
//                    storedParams.push_back(param);
                }
            }
        }
    }
}


// set camera or driver attribute based on string value
bool AlliedCamera::String2Value(tPvHandle aCamera, string paramName, tPvDatatype aType, string & paramValue){

    const char * aLabel = paramName.c_str();
    const char * aValue = paramValue.c_str();

    switch(aType)
    {           
        case ePvDatatypeString:
        {   
            return (PvAttrStringSet(aCamera,aLabel,aValue) == ePvErrSuccess);    
        }
        case ePvDatatypeEnum:
        {            
            return (PvAttrEnumSet(aCamera,aLabel,aValue) == ePvErrSuccess);
        }
        case ePvDatatypeUint32:
        {
            tPvUint32 lValue = atol(aValue);
            tPvUint32 lMin,lMax;
            
           if(PvAttrRangeUint32(aCamera,aLabel,&lMin,&lMax) == ePvErrSuccess)
           {
               if(lMin > lValue)
                   lValue = lMin;
               else
               if(lMax < lValue)
                   lValue = lMax;
                                        
               return PvAttrUint32Set(aCamera,aLabel,lValue) == ePvErrSuccess;
           }
           else
               return false;
        }
        case ePvDatatypeFloat32:
        {
            tPvFloat32 lValue = (tPvFloat32)atof(aValue);
            tPvFloat32 lMin,lMax;
            
           if(PvAttrRangeFloat32(aCamera,aLabel,&lMin,&lMax) == ePvErrSuccess)
           {
                if(lMin > lValue)
                   lValue = lMin;
                else
                if(lMax < lValue)
                   lValue = lMax;            
            
                return PvAttrFloat32Set(aCamera,aLabel,lValue) == ePvErrSuccess;
           }
           else
               return false;
        }
        case ePvDatatypeBoolean:
        {
            if(!(strcmp(aValue,"true")))
                return PvAttrBooleanSet(aCamera,aLabel,true) == ePvErrSuccess;
            else
                return PvAttrBooleanSet(aCamera,aLabel,false) == ePvErrSuccess;
        }
        default:
            return false;
    }       
}

// encode the value of a given attribute in a string
bool AlliedCamera::Value2String(tPvHandle aCamera, string paramName,tPvDatatype aType, string & paramValue)
{

    const char * aLabel = paramName.c_str();
    stringstream stream;
    
    switch(aType)
    {           
        case ePvDatatypeString:
        {
            char aString[1024];
            memset(aString, 0, 1024);
            unsigned long aLength;

            if( PvAttrStringGet(aCamera,aLabel,aString,aLength,NULL) == ePvErrSuccess){
                paramValue = aString;
                return true;
            }else{
                return false;
            }
        }
        case ePvDatatypeEnum:
        {
            char aString[1024];
            memset(aString, 0, 1024);
            unsigned long aLength;
            
            if( PvAttrEnumGet(aCamera,aLabel,aString,aLength,NULL) == ePvErrSuccess){
                paramValue = aString;
                return true;
            }else{
                return false;
            }
        }
        case ePvDatatypeUint32:
        {
            tPvUint32 lValue;
            
            if(PvAttrUint32Get(aCamera,aLabel,&lValue) == ePvErrSuccess)
            {
                stream << lValue;
                paramValue = stream.str();
                return true;
            }
            else
                return false;
            
        }
        case ePvDatatypeFloat32:
        {
            tPvFloat32 lValue;
            
            if(PvAttrFloat32Get(aCamera,aLabel,&lValue) == ePvErrSuccess)
            {
                stream << lValue;
                paramValue = stream.str();
                return true;
            }
            else
                return false;
        }
        case ePvDatatypeBoolean:
        {
            tPvBoolean lValue;
            
            if(PvAttrBooleanGet(aCamera,aLabel,&lValue) == ePvErrSuccess)
            {
                if(lValue)
                    paramValue = "true";
                else
                    paramValue = "false";
                return true;
            }
            else
                return false;            
        }        
        default:
            return false;
    }        
}


//--------------------------------------------------------------
// allocate memory
// return value: true == success, false == fail
bool AlliedCamera::CameraSetup(){
    cout << " AlliedCamera::CameraSetup " << endl;
    tPvErr errCode;
	unsigned long FrameSize = 0;

	// Calculate frame buffer size
    if((errCode = PvAttrUint32Get(Handle,"TotalBytesPerFrame",&FrameSize)) != ePvErrSuccess){
		printf("Cam[%u]: Get TotalBytesPerFrame err: %u\n", info.deviceNo, errCode);
		return false;
	}
    
    Frames.clear();
    for(int i = 0; i < 3; i++){
        Frames.push_back(new tPvFrame());
        Frames[i]->ImageBuffer = new char[FrameSize];
        Frames[i]->ImageBufferSize = FrameSize;
    }

	return true;
}

//--------------------------------------------------------------
void AlliedCamera::CameraUnsetup(){
    cout << " AlliedCamera::CameraUnsetup " << endl;

    tPvErr errCode;
	
    if((errCode = PvCameraClose(Handle)) != ePvErrSuccess){
		printf("Cam[%u]: PvCameraClose err: %u\n", info.deviceNo, errCode);
	}
	else{
		printf("Cam[%u]: Closed.\n", info.deviceNo);
	}

    Handle = NULL;
    reset();
}

//--------------------------------------------------------------
// setup and start streaming with PvCaptureWaitForFrameDone
// return value: true == success (code exits on Ctrl+C abort), false == fail (other failure)
bool AlliedCamera::CameraStart(){
    cout << " AlliedCamera::CameraStart " << endl;

    tPvErr errCode;
	bool failed = false;
	int Index = 0;
    unsigned long Last = 0;

    // NOTE: This call sets camera PacketSize to largest sized test packet, up to 8228, that doesn't fail
	// on network card. Some MS VISTA network card drivers become unresponsive if test packet fails. 
	// Use PvUint32Set(handle, "PacketSize", MaxAllowablePacketSize) instead. See network card properties
	// for max allowable PacketSize/MTU/JumboFrameSize. 
	if((errCode = PvCaptureAdjustPacketSize(Handle,8228)) != ePvErrSuccess){
		printf("Cam[%u]: PvCaptureAdjustPacketSize err: %u\n", info.deviceNo, errCode);
		return false;
	}

//	// Assuming hardware setup is multiple cameras on a switch, with a single GigE NIC
//	// Therefore total available bandwidth is ~ 115MB. Divide bandwidth amongst cameras
//	// or get packet collisions, i.e. frames returning with ePvErrDataMissing.
//	if((errCode = PvAttrUint32Set(Handle,"StreamBytesPerSecond", 115000000 / maxNumberCameras )) != ePvErrSuccess)
//	{
//		printf("Cam[%u]: Set StreamBytesPerSecond err: %u\n", info.deviceNo, errCode);
//		return false;
//	}

    // start driver capture stream 
	if((errCode = PvCaptureStart(Handle)) != ePvErrSuccess)
	{
		printf("Cam[%u]: PvCaptureStart err: %u\n", info.deviceNo, errCode);
		return false;
	}
	
    ofSleepMillis(1000);
    
	// queue frames. No FrameDoneCB callback function. 
	for(int i=0;i < Frames.size() && !failed; i++)
	{           
		if((errCode = PvCaptureQueueFrame(Handle,Frames[i],NULL)) != ePvErrSuccess)
		{
			printf("Cam[%u]: PvCaptureQueueFrame err: %u\n", info.deviceNo, errCode);
			failed = true;
		}
	}

	if (failed){
		return false;
    }
	
    // set the camera in freerun trigger, continuous mode, and start camera receiving triggers
	if((PvAttrEnumSet(Handle,"FrameStartTriggerMode","Freerun") != ePvErrSuccess) ||
		(PvAttrEnumSet(Handle,"AcquisitionMode","Continuous") != ePvErrSuccess) ||
		(PvCommandRun(Handle,"AcquisitionStart") != ePvErrSuccess))
	{		
		printf("Cam[%u]: failed to set camera attributes\n", info.deviceNo);
		return false;
	}	 
    
	//Loop until failure or CRTL+C abort
    while(!failed && isThreadRunning())
    {
        //cout << " waiting for frame " << endl;
        
		//Wait for [Index] frame of ALLIED_FRAME_COUNT num frames
		//wait for frame to return from camera to host
        while(isThreadRunning() && (errCode = PvCaptureWaitForFrameDone(Handle,Frames[Index],2000)) == ePvErrTimeout){
			printf("\nCam[%u]: Waiting for frame to return to host...\n", info.deviceNo);
        }
	
		if(errCode != ePvErrSuccess)
        { 
			//likely camera unplugged
			printf("Cam[%u]: PvCaptureWaitForFrameDone err: %u\n", info.deviceNo, errCode);
			failed = true;
		}
		else
		{
        
            ofImageType imageType = OF_IMAGE_UNDEFINED;
            int frameImageType = Frames[Index]->Format;
            
            if( frameImageType == ePvFmtMono8 ){
                imageType = OF_IMAGE_GRAYSCALE;
            }else if( frameImageType == ePvFmtRgb24 || frameImageType == ePvFmtBgr24 ){
                imageType = OF_IMAGE_COLOR;
            }else if( frameImageType == ePvFmtRgba32 || frameImageType == ePvFmtBgra32){
                imageType = OF_IMAGE_COLOR_ALPHA;
            }else{
                ofLogError("AlliedCamera::CameraStart") << " Image frame format not supported - add implementation / comversion here ";
            }
            
            if( imageType != OF_IMAGE_UNDEFINED ){
                //cout << " copying data " << endl;
                
                if( lock() ){
                    if( bFrameNewInteral == false ){
                        unsigned char * buffer = (unsigned char *)Frames[Index]->ImageBuffer;
                        pixelsInternal.setFromPixels(buffer, Frames[Index]->Width, Frames[Index]->Height, imageType);
                        bFrameNewInteral = true;
                    }
                    unlock();
                }
            }
        
            //cout << " got frame " << endl;

			// if frame hasn't been cancelled, requeue frame
			if(Frames[Index]->Status != ePvErrCancelled)
			{
				//Check for gaps in FrameCount due to image returning from camera with no frame queued.
				//This should never happen, as we use a multiple frame circular buffer. 
				if(Last + 1 != Frames[Index]->FrameCount)
				{
					//Note missing frame
					Discarded++;
				}
                    
				Last = Frames[Index]->FrameCount;
                
				//Requeue [Index] frame of ALLIED_FRAME_COUNT num frames
				if ((errCode = PvCaptureQueueFrame(Handle,Frames[Index],NULL)) != ePvErrSuccess)
				{
					printf("Cam[%u]: PvCaptureQueueFrame err %u\n", info.deviceNo, errCode);
					failed = true;
				}
				
				//Increment [Index]
				Index++;
				if(Index >= Frames.size()){
					Index = 0;
                }
			}
			else
			{
				//Cancelled
				failed = true;
			}
		}
    }

	return !failed;
}

// stop streaming, clear queue.
void AlliedCamera::CameraStop(){
    cout << " AlliedCamera::CameraStop " << endl;

    tPvErr errCode;
	
	//stop camera receiving triggers
	if ((errCode = PvCommandRun(Handle,"AcquisitionStop")) != ePvErrSuccess){
		printf("Cam[%u]: AcquisitionStop err: %u\n", info.deviceNo, errCode);
    }
    else{
		printf("Cam[%u]: AcquisitionStop success.\n", info.deviceNo);
    }
    
	//PvCaptureQueueClear aborts any actively written frame with Frame.Status = ePvErrDataMissing
	//Further queued frames returned with Frame.Status = ePvErrCancelled
	
	//Add delay between AcquisitionStop and PvCaptureQueueClear
	//to give actively written frame time to complete
	ofSleepMillis(200);
	
	printf("Cam[%u]: Calling PvCaptureQueueClear...\n", info.deviceNo);
	if ((errCode = PvCaptureQueueClear(Handle)) != ePvErrSuccess)
		printf("Cam[%u]: PvCaptureQueueClear err: %u\n", info.deviceNo, errCode);
	else
		printf("Cam[%u]: ...Queue cleared.\n", info.deviceNo);  

	//stop driver stream
	if ((errCode = PvCaptureEnd(Handle)) != ePvErrSuccess)
		printf("Cam[%u]: PvCaptureEnd err: %u\n", info.deviceNo, errCode);
	else
		printf("Cam[%u]: Driver stream stopped.\n", info.deviceNo);
}


//--------------------------------------------------------------
void AlliedCamera::threadedFunction(){

    cout << " AlliedCamera::threadedFunction trying to open camera " << info.deviceNo << " named: " << info.deviceName << " with ip: " << info.ipAddress << endl;

    tPvErr errCode;
    
    if((errCode = PvCameraOpenByAddr(info.ipAddr,ePvAccessMaster,&Handle)) == ePvErrSuccess){

        char IP[128];
        char Name[128];

        // Read the IP and Name strings
        if(((errCode = PvAttrStringGet(Handle,"DeviceIPAddress",IP,128,NULL)) == ePvErrSuccess) &&
           ((errCode = PvAttrStringGet(Handle,"CameraName",Name,128,NULL)) == ePvErrSuccess)){
            
            ofLogNotice("AlliedCamera::openCamera") << " opened camera " << info.deviceNo << " with ip " << info.ipAddress << endl;
            if(CameraSetup()){
                CameraStart();
                printf("Cam[%u]: %lu frames missed by host frame queue\n", info.deviceNo, Discarded);
                CameraStop();
                CameraUnsetup();
            }
        }else{
            ofLogError() << " unable to open camera " << Name << " with IP " << IP << endl;
        }
        
        if( Handle ){
            PvCameraClose(Handle);
            Handle = NULL;
        }
    }else{
            ofLogError() << " unable to open camera by inet_addr " << info.ipAddr << " error is " << errCode <<  endl;
    }
    
    stopThread();
}



#pragma mark - Static Methods


//--------------------------------------------------------------
void AlliedCamera::initApi(){
    if( initCount == 0 ){
        PvInitialize();
    }
    initCount++;
}

//--------------------------------------------------------------
void AlliedCamera::releaseApi(){
    int nextcount = initCount-1;
    if( initCount == 1 && nextcount == 0 ){
        PvUnInitialize();
    }
    initCount = nextcount;
    if( initCount < 0){
        initCount = 0;
    }
}

//--------------------------------------------------------------
void AlliedCamera::addCallbacks() {
    if( !bHasLinkListeners ) {
        // register the callback for the link events
        PvLinkCallbackRegister( AlliedPvLinkListener::LinkEventCB, ePvLinkAdd, &listener );
        PvLinkCallbackRegister( AlliedPvLinkListener::LinkEventCB, ePvLinkRemove, &listener );
    }
    bHasLinkListeners = true;
}

//--------------------------------------------------------------
void AlliedCamera::removeCallbacks() {
    if( bHasLinkListeners ) {
        PvLinkCallbackUnRegister( AlliedPvLinkListener::LinkEventCB, ePvLinkAdd);
        PvLinkCallbackUnRegister( AlliedPvLinkListener::LinkEventCB, ePvLinkRemove);
    }
    bHasLinkListeners = false;
}

//--------------------------------------------------------------
vector < AlliedCamera::CameraInfo > AlliedCamera::ListDevices(){
    AlliedCamera::initApi();
    
    ofSleepMillis(200);
    
    tPvCameraInfo   cameraList[MAX_CAMERA_LIST];
    unsigned long   cameraNum = 0;
    unsigned long   cameraRle;

    vector < CameraInfo > info;
    
    int checkCount = 3;
    while(checkCount > 0)
    {
        info.clear();
        checkCount--; 

        // first, get list of reachable cameras.
        cameraNum = PvCameraList(cameraList,MAX_CAMERA_LIST,NULL);

       // keep how many cameras listed are reachable
        cameraRle = cameraNum;    

        // then we append the list of unreachable cameras.
        if (cameraNum < MAX_CAMERA_LIST)
        {
            cameraNum += PvCameraListUnreachable(&cameraList[cameraNum],
                                                 MAX_CAMERA_LIST-cameraNum,
                                                 NULL);
        }

        if(cameraNum)
        {
            struct in_addr addr;
            tPvIpSettings Conf;
            tPvErr lErr;
            
            // display info
            for(unsigned long i=0;i<cameraNum;i++){
            
                AlliedCamera::CameraInfo camInfo;
                camInfo.serialNo    = cameraList[i].SerialString;
                camInfo.deviceName  = cameraList[i].DisplayName;
                camInfo.hardwareID  = ofToString(cameraList[i].UniqueId);
                camInfo.UniqueId    = cameraList[i].UniqueId;
                camInfo.deviceNo    = i;
                
                if(i < cameraRle)
                {
                    // get the camera's IP configuration
                    int IPGrabCount = 10;
                    
                    bool bSuccess = false;
                    
                    while(bSuccess == false && IPGrabCount > 0 ){

                        if((lErr = PvCameraIpSettingsGet(cameraList[i].UniqueId,&Conf)) == ePvErrSuccess){
                            addr.s_addr = Conf.CurrentIpAddress;
                            

    //                        printf("%s - %8s - Unique ID = % 8lu IP@ = %15s [%s]\n",cameraList[i].SerialString,
    //                                                                    cameraList[i].DisplayName,
    //                                                                    cameraList[i].UniqueId,
    //                                                                    inet_ntoa(addr),
    //                                                                    cameraList[i].PermittedAccess & ePvAccessMaster ? "available" : "in use");

                            camInfo.available = true;
                            camInfo.ipAddress = inet_ntoa(addr);
                            camInfo.ipAddr  = Conf.CurrentIpAddress;
                            bSuccess = true;
                        }
                        
                        IPGrabCount--;
                        ofSleepMillis(10);
                    }
                    camInfo.available = bSuccess;
                    
                }
                else{
                    camInfo.available = false;
                }
                
                info.push_back(camInfo);
            }
            
        }
        
        ofSleepMillis(100);
    }
    
    for(int i = 0; i < info.size(); i++){
        ofLogNotice("AlliedCamera::ListDevices") << "[" << info[i].deviceNo << "] " << info[i].serialNo << " " << info[i].deviceName << " << unique id: " << cameraList[i].UniqueId << " IP address " << info[i].ipAddress << " Available?: " << info[i].available << endl;
    }
    
    if( info.size() == 0 ){
        ofLogWarning("AlliedCamera::ListDevices") << " no devices found " << endl;
    }
    
    AlliedCamera::releaseApi();
    
    return info;
}


void AlliedCamera::setMaxSimultaniousCameras(int maxNoCameras){
    maxNumberCameras = maxNoCameras;
}

int AlliedCamera::getMaxSimultaniousCameras() {
    return maxNumberCameras;
}


//void AlliedCamera::LinkEventCB(void* aContext,tPvInterface aInterface,tPvLinkEvent aEvent,unsigned long aUniqueId) {
//    cout << "LinkEventCB :: id : " << aUniqueId << endl;
//    CApplication* lApplication = (CApplication*)aContext;
//    
//    if(lApplication->iMain)
//    {
//        // append a custom event
//        CLinkEvent lEvent(aEvent,aUniqueId);
//        lEvent.SetEventObject(lApplication);
//        ::wxPostEvent(lApplication->iMain->GetEventHandler(),lEvent);
//    }
//}










