//
//  AlliedCamera.h
//  emptyExample
//
//  Created by Theodore Watson on 9/30/14.
//
//

#pragma once
#include "ofMain.h"
#include <PvApi.h>

#define ALLIED_FRAME_COUNT 3

class AlliedLinkEvent {
public:
    void* Context;
    tPvInterface Interface;
    tPvLinkEvent Event;
    unsigned long UniqueId;
};


class AlliedPvLinkListener {
public:
    static ofEvent< AlliedLinkEvent > LinkEvent;
    
    static void LinkEventCB( void* aContext, tPvInterface aInterface, tPvLinkEvent aEvent, unsigned long aUniqueId ) {
//        cout << "PvLinkListener :: event : " << aEvent << " id: " << aUniqueId << endl;
        AlliedLinkEvent newEvent;
        newEvent.Context = aContext;
        newEvent.Interface = aInterface;
        newEvent.UniqueId = aUniqueId;
        ofNotifyEvent( LinkEvent, newEvent );
    }
};



class AlliedCamera : public ofThread{
    public:
    
    class CameraInfo{
        public:
            CameraInfo(){
                deviceNo = -1;
                available = false;
                ipAddr = 0;
                UniqueId = -1;
            }
        
            string deviceName;
            string hardwareID;
            string ipAddress;
            unsigned long ipAddr;
            string serialNo;
            bool available;
            int deviceNo;
            unsigned long UniqueId;
    };
    
    class CameraParam{
        friend class AlliedCamera;
        
        public:
            CameraParam(){
                valueF = 0.0;
                minF = 0.0;
                maxF = 0.0;
                valueB = false;
                writable = false;
                type = ePvDatatypeUnknown;
                valueL = 0;
                minL = 0;
                maxL = 0;
                enumSelected = 0; 
            }
        
            string getName(){
                return name;
            }
        
            tPvDatatype getType(){
                return type;
            }
        
            float getValueF(){
                return valueF;
            }
        
            float getMinValueF(){
                return minF;
            }

            float getMaxValueF(){
                return maxF;
            }
        
            unsigned long getValueUL(){
                return valueL;
            }

            float getMinValueUL(){
                return minL;
            }

            float getMaxValueUL(){
                return maxL;
            }
        
            bool getValueB(){
                return valueB;
            }
        
            string getValueS(){
                return valueS;
            }
        
            string getValueEnum(){
                return valueS;
            }
        
            vector <string> getEnumValues(){
                return enumVals;
            }
        
            unsigned int getEnumSelected(){
                return enumSelected;
            }
        
            bool isWritable(){
                return writable;
            }
        
        protected:
            string name;

            float valueF;
            float minF;
            float maxF;
        
            unsigned long valueL;
            unsigned long minL;
            unsigned long maxL;

            tPvDatatype type;
            bool writable;
            bool valueB;
        
            string valueS;
            vector <string> enumVals;
            unsigned int enumSelected;
    };

    //STATIC METHODS
    static vector < CameraInfo > ListDevices();
    static void initApi();
    static void releaseApi();
    static void addCallbacks();
    static void removeCallbacks();
    static void setMaxSimultaniousCameras(int maxNoCameras);
    static int getMaxSimultaniousCameras();
    
//    static void LinkEventCB( void* aContext, tPvInterface aInterface, tPvLinkEvent aEvent, unsigned long aUniqueId );
    
    AlliedCamera();
    ~AlliedCamera();
    bool openCamera(unsigned int deviceID, bool waitForFirstFrame);
    bool openCamera(string hardwareID, bool waitForFirstFrame);
    void closeCamera();
    bool isCameraRunning(); 

    void update();
    bool isFrameNew();
    const ofPixels & getPixelsRef();
    
    bool setValueF(string paramName, float value);
    bool setValueUL(string paramName, unsigned long value);
    bool setValueB(string paramName, bool value);
    bool setValueS(string paramName, string value);
    bool setValueEnum(string paramName, string value);

    bool getValueF(string paramName, float & value);
    bool getValueUL(string paramName, unsigned long & value);
    bool getValueB(string paramName, bool & value);
    bool getValueS(string paramName, string & value);
    bool getValueEnum(string paramName, string & value);
    
    CameraInfo getCameraInfo(){
        return info;
    }
    
    vector <CameraParam> getCameraParameters();
    
    
    ofPixels pixels;
    protected:
        void startThreadedCapture(bool bWaitForFirstFrame);
        void reset();
        void threadedFunction();

        bool CameraSetup();
        bool CameraStart();
        void CameraStop();
        void CameraUnsetup();
    
    
        bool String2Value(tPvHandle aCamera, string paramName, tPvDatatype aType, string & paramValue);
        bool Value2String(tPvHandle aCamera, string paramName, tPvDatatype aType, string & paramValue);
    
        bool buildAttributeList();
        void QueryAttribute(const char* aLabel);

        CameraInfo info;
        tPvHandle Handle = NULL;
        vector <tPvFrame *> Frames;
        unsigned long Discarded = 0;

        bool bFrameNew = false;
        bool bFrameNewInteral = false;
    
    
        ofPixels pixelsInternal;
    
        ofParameterGroup params;
        vector < ofAbstractParameter * > storedParams;

        vector <CameraParam> camParams;
    
    
};



